"use strict"

/****** CHECK EXTERNAL LIBRARIES LOADED ******/
const leafletLoaded = typeof L !== 'undefined'
const d3Loaded = typeof d3 !== 'undefined'
if (!leafletLoaded) {
  console.log('[geogrid.js] Leaflet needs to be loaded first / only pure functions available')
  window.L = {}
}
if (!d3Loaded) console.log('[geogrid.js] D3.js needs to be loaded first / only pure functions available')

/****** IMPORTS ******/
import './geogrid.scss'
import {defaultOptions, initCore} from './geogrid.core.js'
import {Download} from './geogrid.download.js'
import {RendererSVG} from './geogrid.rendererSVG.js'
import {RendererWebGL} from './geogrid.rendererWebGL.js'

/****** PURE FUNCTION ******/
L.isea3hToGeoJSON = (options, callback) => {
  // set options
  options = Object.assign({}, () => {}, defaultOptions, options)
  // init core
  initCore(options, d => {}, callback, false)
}

/****** PLUGIN ******/
if (leafletLoaded && d3Loaded) L.ISEA3HLayerPlugin = class ISEA3HLayerPlugin {
  onAdd(layer) {
    this._layer = layer
  }
  downloadData() {
    this._layer._updateData()
  }
  render() {
    this._layer._render()
  }
  neighbors(cell, callback) {
    return this._layer._neighbors(cell, callback)
  }
  getParameter(parameter) {
    return this._layer.options.parameters[parameter]
  }
  setParameter(parameter, value) {
    this._layer.options.parameters[parameter] = value
  }
  setCellColor(cell, value) {
    if (value) this._layer._data.overwriteColor(cell.id, value)
    else this._layer._data.overwriteColor(cell.id, null)
  }
  resetCellColor() {
    this._layer._data.resetOverwriteColor()
  }
  setCellSize(cell, value) {
    if (value) this._layer._data.overwriteSize(cell.id, value)
    else this._layer._data.overwriteSize(cell.id, null)
  }
  resetCellSize() {
    this._layer._data.resetOverwriteSize()
  }
  setCellContourColor(cell, value) {
    if (value) this._layer._data.overwriteContourColor(cell.id, value)
    else this._layer._data.overwriteContourColor(cell.id, null)
  }
  resetCellContourColor() {
    this._layer._data.resetOverwriteContourColor()
  }
  setCellContourWidth(cell, value) {
    if (value) this._layer._data.overwriteContourWidth(cell.id, value)
    else this._layer._data.overwriteContourWidth(cell.id, null)
  }
  resetCellContourWidth() {
    this._layer._data.resetOverwriteContourWidth()
  }
}

/****** LAYER ******/
if (leafletLoaded && d3Loaded) L.ISEA3HLayer = L.Layer.extend({
  options: defaultOptions,
  initialize: function(options) {
    this._initialized = false
    this._map = null

    // init options I
    L.Util.setOptions(this, options)

    // event listener for web worker
    const eventListener = d => {
      if (this._map) switch (d.task) {
        case 'resultPluginsHover':
          if (!d.cell || !this._hoveredCells.map(c => c.idLong).includes(d.cell.idLong)) {
            for (const cell of this._hoveredCells) {
              const ePlugin = eventForPlugin(cell)
              if (cell.idLong !== d.cell.idLong) for (let p of this._plugins) if (p.onUnhover !== undefined) p.onUnhover(ePlugin)
            }
            if (d.cell) {
              const ePlugin = eventForPlugin(d.cell)
              this._hoveredCells = [d.cell]
              for (let p of this._plugins) if (p.onHover !== undefined) p.onHover(ePlugin)
            }
          }
          break
        case 'resultPluginsClick':
          const ePlugin = eventForPlugin(d.cell)
          for (let p of this._plugins) if (p.onClick !== undefined) p.onClick(ePlugin)
          break
        case 'resultFindNeighbors':
          this._execPluginCallback(d.uid, d.neighbors === null ? null : d.neighbors.map(cell => eventForPlugin(cell)))
          break
      }
    }

    // init the core
    const core = initCore(this.options, eventListener, () => this._visualizeData(), true)
    this.options = core.options
    this._progress = core._progress
    this._data = core._data
    this._webWorker = core._webWorker
    this._processDataInWebWorker = core._processDataInWebWorker
    this._webWorkerPostMessage = core._webWorkerPostMessage

    // init plugins
    if (this._plugins === undefined) this._plugins = []
    if (this._pluginCallbacks === undefined) this._pluginCallbacks = {}
    this._hoveredCells = []

    // choose renderer
    if (this.options.renderer.toLowerCase() == 'svg') this._renderer = new RendererSVG(this.options, this._progress, this._data)
    else {
      if (typeof PIXI === 'undefined') this._progress.error('pixi.js needs to be loaded first')
      if (typeof L.pixiOverlay === 'undefined') this._progress.error('Leaflet.PixiOverlay needs to be loaded first')
      this._renderer = new RendererWebGL(this.options, this._progress, this._data)
    }

    // event for plugin
    const eventForPlugin = cell => cell === null || cell === undefined ? null : {
      lat: cell.lat,
      lon: cell.lon,
      cell: cell,
      data: this._data.dataForId(cell.id),
      ...(this.options.dataMap !== null ? {dataNotMapped: this._data.dataForIdNotMapped(cell.id)} : {}),
    }
  },
  onAdd: function(map) {
    if (!map) return
    this._map = map
    this._renderer.add(map)
    this._updateData()

    // plugins
    map.on('mousemove', this._onMouseMove, this)
    map.on('mouseout', this._onMouseOut, this)
    map.on('click', this._onClick, this)
    
    // events
    map.on('viewreset', this._onReset, this)
    map.on('zoomend', this._onReset, this)
    map.on('moveend', this._onReset, this)
  },
  onRemove: function(map) {
    if (!map) return
    this._progress.remove()
    this._renderer.remove(map)
    this._webWorker.terminate()

    // plugins
    map.off('mousemove', this._onMouseMove, this)
    map.off('mouseout', this._onMouseOut, this)
    map.off('click', this._onClick, this)
    
    // events
    map.off('viewreset', this._onReset, this)
    map.off('zoomend', this._onReset, this)
    map.off('moveend', this._onReset, this)

    this._map = null
  },
  update: function(options) {
    let reinitialize = false
    let updateData = false
    let processData = false
    let produceGeoJSON = false
    // messages
    const notYetImplemented = o => {
      console.log(`[WARNING] update of "${o}" not yet implemented; will re-initialize`)
      reinitialize = true
    }
    // check options
    if (options.url != undefined) updateData = true
    if (options.data != undefined) reinitialize = true
    if (options.silent != undefined) notYetImplemented('silent')
    if (options.debug != undefined) notYetImplemented('debug')
    if (options.resolution != undefined) notYetImplemented('resolution')
    if (options.parameters != undefined) updateData = true
    if (options.cellColorKey != undefined) produceGeoJSON = true
    if (options.cellColorMin != undefined) produceGeoJSON = true
    if (options.cellColorMax != undefined) produceGeoJSON = true
    if (options.cellColorScale != undefined) produceGeoJSON = true
    if (options.cellColorNoData != undefined) produceGeoJSON = true
    if (options.cellColorNoKey != undefined) produceGeoJSON = true
    if (options.cellColorOpacity != undefined) produceGeoJSON = true
    if (options.cellSizeKey != undefined) produceGeoJSON = true
    if (options.cellSizeMin != undefined) produceGeoJSON = true
    if (options.cellSizeMax != undefined) produceGeoJSON = true
    if (options.cellSizeScale != undefined) produceGeoJSON = true
    if (options.cellSizeNoData != undefined) produceGeoJSON = true
    if (options.cellSizeNoKey != undefined) produceGeoJSON = true
    if (options.cellContourColor != undefined) produceGeoJSON = true
    if (options.cellContourWidth != undefined) produceGeoJSON = true
    if (options.colorProgressBar != undefined) notYetImplemented('colorProgressBar')
    if (options.colorDebug != undefined) notYetImplemented('colorDebug')
    if (options.colorDebugEmphasized != undefined) notYetImplemented('colorDebugEmphasized')
    if (options.dataKeys != undefined) notYetImplemented('dataKeys')
    if (options.dataMap != undefined) produceGeoJSON = true
    if (options.attribution != undefined) notYetImplemented('attribution')
    if (options.bboxViewPad != undefined) notYetImplemented('bboxViewPad')
    if (options.bboxDataPad != undefined) notYetImplemented('bboxDataPad')
    if (options.renderer != undefined) notYetImplemented('renderer')
    // re-initialize
    if (reinitialize) {
      const map = this._map
      this.onRemove(this._map)
      this.initialize({
        ...this.options,
        ...options,
      })
      this.onAdd(map)
      return
    }
    // copy options
    for (const k in options) this.options[k] = options[k]
    // update data
    if (updateData) {
      this._updateData()
      return
    }
    // process data
    if (processData) {
      this._processData()
      return
    }
    // produce GeoJSON
    if (produceGeoJSON) {
      this._data.produceGeoJSON()
      this._visualizeData()
      return
    }
  },
  addPlugin: function(plugin) {
    plugin.onAdd(this)
    this._plugins.push(plugin)
    if (plugin.onHover !== undefined) this._pluginsOnHover = true
    if (plugin.onClick !== undefined) this._pluginsOnClick = true
  },
  _uuidv4: function() {
    return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, c => {
      let r = Math.random() * 16 | 0, v = c == 'x' ? r : (r & 0x3 | 0x8)
      return v.toString(16)
    })
  },
  _storePluginCallback: function(callback) {
    const uid = this._uuidv4()
    this._pluginCallbacks[uid] = callback
    return uid
  },
  _execPluginCallback: function(uid, ...es) {
    this._pluginCallbacks[uid](...es)
    delete this._pluginCallbacks[uid]
  },
  _neighbors: function(cell, callback) {
    this._webWorkerPostMessage({
      task: 'findNeighbors',
      taskResult: 'resultFindNeighbors',
      uid: this._storePluginCallback(callback),
      idLong: cell.idLong,
    })
  },
  _paddedBounds: function() {
    return this._map.getBounds().pad(this.options.bboxViewPad - 1)
  },
  _updateData: function() {
    const t = this
    // proceed only if data is available
    if (t.options.url === null && t.options.data === null) return
    // download the data
    t._progress.showProgress()
    t._progress.debugStep(t.options.url !== null ? 'download data' : 'update data', 2.5)
    t._bboxData = this._map.getBounds().pad(this.options.bboxDataPad - 1)
    const resolution = t.options.resolution(t._map.getZoom())
    if (t.options.url !== null) new Download(t.options, resolution, t._progress).load(t._bboxData, data => {
      t.options.data = data
      t._resolutionData = resolution
      t.fire('dataDownloaded', {data: t.options.data})
      t._processData()
    })
    else t._processData()
  },
  _processData: function() {
    // process data in web worker
    this._processDataInWebWorker({
      north: this._bboxData.getNorth(),
      south: this._bboxData.getSouth(),
      west: this._bboxData.getWest(),
      east: this._bboxData.getEast(),
    })
  },
  _reduceGeoJSON: function() {
    this._progress.debugStep('reduce GeoJSON for area', 70)
    return this._data.reduceGeoJSON(this._paddedBounds())
  },
  _visualizeData: function() {
    const t = this
    // visualize centroids
    if (t._centroids != null) for (let c of t._centroids) c.remove()
    t._centroids = []
    if (t.options.debug) {
      this._progress.debugStep('visualize centroids', 75)
      for (let d of t._data.getCells()) {
        const circle = L.circle([d.lat, d.lon], {
          color: (d.isPentagon) ? t.options.colorDebugEmphasized : t.options.colorDebug,
          fill: (d.isPentagon) ? t.options.colorDebugEmphasized : t.options.colorDebug,
          radius: 3}
        ).on('mouseover', e => console.debug(e.target._d)).addTo(t._map)
        circle._d = d
        t._centroids.push(circle)
      }
    }
    // visualize cells
    t._render()
    // layer has been initialized
    if (!t._initialized) {
      setTimeout(() => {
        t.fire('loadComplete')
      }, 20)
      t._initialized = true
    }
  },
  _onMouseMove: function(e) {
    if (this._pluginsOnHover && this._initialized) this._webWorkerPostMessage({
      task: 'findCell',
      taskResult: 'resultPluginsHover',
      lat: e.latlng.lat,
      lon: e.latlng.lng,
    })
  },
  _onMouseOut: function(e) {
    if (this._hoveredCells === undefined) return
    for (const cell of this._hoveredCells) {
      const ePlugin = {
        cell: cell,
      }
      for (let p of this._plugins) if (p.onUnhover !== undefined) p.onUnhover(ePlugin)
    }
    this._hoveredCells = []
  },
  _onClick: function(e) {
    if (this._pluginsOnHover && this._initialized) this._webWorkerPostMessage({
      task: 'findCell',
      taskResult: 'resultPluginsClick',
      lat: e.latlng.lat,
      lon: e.latlng.lng,
    })
  },
  _onReset: function(e) {
    if (this._data.getGeoJSON() === null || this._data.getGeoJSON() === undefined) return
    // reset after zooming, panning, etc.
    if ((this._paddedBounds && !this._bboxData.contains(this._paddedBounds())) || (this.options.url && this.options.resolution(this._map.getZoom()) !== this._resolutionData)) this._updateData()
    else {
      const geoJSONreduced = this._reduceGeoJSON()
      if (geoJSONreduced && geoJSONreduced.features.length) this._renderer.render(geoJSONreduced)
    }
  },
  _render: function() {
    this._renderer.render(this._reduceGeoJSON())
  },
})

if (leafletLoaded && d3Loaded) L.isea3hLayer = options => new L.ISEA3HLayer(options)
